import numpy as np
import cv2 as cv
import operator
import math

def drawEdge(im, edge, color1, color2):

	for i in range(len(edge)-1):
	    cv.line(im,(edge[i][0],edge[i][1]),(edge[i+1][0],edge[i+1][1]),color1,1,cv.LINE_AA)
	    cv.circle(im,(edge[i][0],edge[i][1]),2,color2,-1)
	cv.circle(im,(edge[-1][0],edge[-1][1]),2,color2,-1)

	return im

def removeSmallContours(cnt, thr):

	temp = []

	for i in range(len(cnt)):
		if cv.contourArea(cnt[i]) > thr:
			temp.append(cnt[i])

	return temp

def calcDist(cnt, dist):

	for i in range(0,len(cnt[0])):
	    mind = 1e9
	    for j in range(0,len(cnt[1])):
	        temp = np.linalg.norm(cnt[0][i]-cnt[1][j])
	        if temp < mind:
	            mind = temp
	            dist[0][i] = j
                
	for i in range(0,len(cnt[1])):
	    mind = 1e9
	    for j in range(0,len(cnt[0])):
	        temp = np.linalg.norm(cnt[1][i]-cnt[0][j])
	        if temp < mind:
	            mind = temp
	            dist[1][i] = j

	return dist 

def merge(cnt, pairs):

	tempt = []
	tempb = []

	###Open the bottom contour###
	x_r = 0
	x_l = 128
	idx_r = 0
	idx_l = 0

	for pair in pairs: 
	    if cnt[0][pair[0]][0,0] > x_r:
	        x_r = cnt[0][pair[0]][0,0]
	        idx_r = pair[0]
	    if cnt[0][pair[0]][0,0] < x_l:
	        x_l = cnt[0][pair[0]][0,0]
	        idx_l = pair[0]

	if idx_r > idx_l:
		tempb = cnt[0][idx_l:idx_r]

	if idx_r < idx_l:
		tempb = np.concatenate((cnt[0][idx_l:len(cnt[0])],cnt[0][0:idx_r+1]),axis=0)

	###Open the top contour###
	x_r = 0
	x_l = 128

	for pair in pairs: 
		if cnt[1][pair[1]][0,0] > x_r:
		    x_r = cnt[1][pair[1]][0,0]
		    idx_r = pair[1]
		if cnt[1][pair[1]][0,0] < x_l:
			x_l = cnt[1][pair[1]][0,0]
			idx_l = pair[1]

	if idx_r > idx_l:
		tempt = np.concatenate((cnt[1][idx_r:len(cnt[1])],cnt[1][0:idx_l+1]),axis=0)

	if idx_r < idx_l:
		tempt = cnt[1][idx_r:idx_l]

	###Merge the contours###
	if (len(tempb) > 0) and (len(tempt)) > 0:
		return [np.concatenate((tempb,tempt),axis=0)]
	else:
		if cv.contourArea(cnt[1]) > cv.contourArea(cnt[0]):
			return cnt[1]
		else: 
			return cnt[0]

def mergeContours(cnt):

	dist = [None,None]
	dist[0] = [None]*len(cnt[0])
	dist[1] = [None]*len(cnt[1])

	pairs = []
	

	###Calculate distances from each point in the contours to all points in the other contour###
	dist = calcDist(cnt, dist)

	###If a pair exists which are the shortest distance to each other, put them in a list###
	for i in range(len(dist[0])):
		if dist[1][dist[0][i]] == i:
			pairs.append([dist[1][dist[0][i]],dist[0][i]])

    ###If more than one pair is in the list of pairs, merge the contours###
	if len(pairs) > 1:
		return merge(cnt, pairs)

    ###If only one pair exist in the list, remove the bottom contour###
	else: 
		if cv.contourArea(cnt[1]) > cv.contourArea(cnt[0]):
			return [cnt[1]]
		else:
			return [cnt[0]]

def findIntersections(cnt, a1, b1):
	intersections = []
	tempInter = []

	for i in range(len(cnt)-1):
		### If a line segment is not vertical, solve using standard method###
		if cnt[i][0]-cnt[i+1][0] != 0:
            
			a2 = (cnt[i][1] - cnt[i+1][1])/(cnt[i][0] - cnt[i+1][0])
			b2 = cnt[i][1] - a2*cnt[i][0]
            
            
            ###if test line and line segment is parrallel, continue###
			if a1 != a2:
				cx = int(round((b2-b1)/(a1-a2),0))
				cy = int(round(b1,0))

			    ###Check if intersections are on the line segment###
				if ( cx >= min([cnt[i][0],cnt[i+1][0]]) ) and ( cx <= max([cnt[i][0],cnt[i+1][0]]) ) and ( cy >= min([cnt[i][1],cnt[i+1][1]]) ) and ( cy <= max([cnt[i][1],cnt[i+1][1]]) ):
					intersections.append([cx,cy])
        
        ###If a line segment is vertical, use x-coord from line segment, and y-coord from test line as intersection##
		else:
 
			cx = cnt[i][0]
			cy = b1

            ###Check if intersections are on the line segment###
			if ( cy >= min([cnt[i][1],cnt[i+1][1]]) ) and ( cy <= max([cnt[i][1],cnt[i+1][1]]) ):                
				intersections.append([cx,cy])


	### Check for the last pair of points to close the contour

	### If a line segment is not vertical, solve using standard method###
	if cnt[-1][0]-cnt[0][0] != 0:
        
		a2 = (cnt[-1][1] - cnt[0][1])/(cnt[-1][0] - cnt[0][0])
		b2 = cnt[-1][1] - a2*cnt[-1][0]
        
        
        ###if test line and line segment is parrallel, continue###
		if a1 != a2:
			cx = int(round((b2-b1)/(a1-a2),0))
			cy = int(round(b1,0))

		    ###Check if intersections are on the line segment###
			if ( cx >= min([cnt[-1][0],cnt[0][0]]) ) and ( cx <= max([cnt[-1][0],cnt[0][0]]) ) and ( cy >= min([cnt[-1][1],cnt[0][1]]) ) and ( cy <= max([cnt[-1][1],cnt[0][1]]) ):
				intersections.append([cx,cy])
    
    ###If a line segment is vertical, use x-coord from line segment, and y-coord from test line as intersection##
	else:
        
		cx = cnt[-1][0]
		cy = b1
		         
        ###Check if intersections are on the line segment###
		if ( cy >= min([cnt[-1][1],cnt[0][1]]) ) and ( cy <= max([cnt[-1][1],cnt[0][1]]) ):                
			intersections.append([cx,cy])


	for intersection in intersections:
		if intersection not in tempInter:
			tempInter.append(intersection) 

	return sorted(tempInter,key=operator.itemgetter(0))

def cleanupPts(left, right, img):

	### Check if points are on the edge of camera FOV ###
	temp = []
	tempRemoved = []
	for i in range(len(right)):
		if right[i][0] < 125:
			if img[right[i][1],right[i][0]+3] != 0:
				temp.append(right[i])
			else:
				tempRemoved.append(right[i])
		else: 
			temp.append(right[i])

	if len(tempRemoved) > 0:
		tempRemoved = sorted(tempRemoved, key=operator.itemgetter(1))
		temp.append(tempRemoved[0])
	right = temp

	temp = []
	tempRemoved = []
	for i in range(len(left)):
		if left[i][0] > 2:
			if img[left[i][1],left[i][0]-3] != 0:
				temp.append(left[i])
			else:
				tempRemoved.append(left[i])
		else:
			temp.append(left[i])

	if len(tempRemoved) > 0:
		tempRemoved = sorted(tempRemoved, key=operator.itemgetter(1))
		temp.append(tempRemoved[0])
			
	left = temp
	
	### Remove points on image edge ###
	temp = []
	for i in range(len(right)):
	    if (right[i][0] > 0) and (right[i][0] < 127): 
	        temp.append(right[i])	
	
	right = temp

	temp = []
	for i in range(len(left)):
	    
	    if (left[i][0] > 0) and (left[i][0] < 127): 
	        temp.append(left[i])
			
	left = temp

	return left, right	

def spacePoints(pts):
	minDist = 5
	### Remove points too close to other points
	pts = list(reversed(sorted(pts, key=operator.itemgetter(1))))

	temp = []
	if len(pts) > 0:
		temp.append(pts[0])

		for i in range(1,len(pts)):
			dist = math.sqrt((pts[i][0] -pts[i-1][0])**2 + (pts[i][1] -pts[i-1][1])**2)
			if dist > minDist:
				temp.append(pts[i])

		pts = temp


	return np.array(pts)

def warpPoints(pts, warpMultiplier):

	h = np.load('homography.npy')

	for i in range(len(pts)):
	    temp = np.matmul(h,[[pts[i][0]],[pts[i][1]],[1]])
	    temp = temp/(temp[2]*warpMultiplier)
	    pts[i] = temp[:2]

	pts = [x.astype(int) for x in pts]

	temp = []

	for i in range(len(pts)):
		if (pts[i][0] > 0) and (pts[i][0] < 128) and (pts[i][1] > 0) and (pts[i][1] < 128):
			temp.append(pts[i])

	return np.squeeze(temp)

def splitLR(cnt, mask, img):

	left = []
	right = []

	###Generate points on left and right edge of contour###
	for i in range(len(cnt)):

		###Make horizontal test line in the y-coordinate of a corner in the contour###
		a1 = 0
		b1 = cnt[i][1]

		intersections = findIntersections(cnt,a1,b1)

		###If 2 intersections, append to each edge ###
		if len(intersections) == 2:
	        
			left.append(intersections[0])
			right.append(intersections[1])
	    
	    ###If 3 intersections, first check for zero values between and sort accordingly, or take the pair with the largest distance between them###
		if len(intersections) == 3:
			if mask[intersections[0][1]][int((intersections[1][0]+intersections[0][0])/2)] == 0:
				left.append(intersections[1])
				right.append(intersections[2])
			elif mask[intersections[0][1]][int((intersections[1][0]+intersections[2][0])/2)] == 0:
				left.append(intersections[0])
				right.append(intersections[1])
			elif (intersections[1][0]-intersections[0][0]) > (intersections[2][0]-intersections[1][0]):
				left.append(intersections[0])
				right.append(intersections[1])
			else:
				left.append(intersections[1])
				right.append(intersections[2])
	        
	    ###If 4 intersections, chose the pair with largest distance between them
		if len(intersections) == 4:
			if (intersections[1][0]-intersections[0][0]) > (intersections[3][0]-intersections[2][0]):
				left.append(intersections[0])
				right.append(intersections[1])
			else:
				left.append(intersections[2])
				right.append(intersections[3])

    
	left, right = cleanupPts(left, right, img)

	### Sort edge points by y-coordinates ###         
	right = sorted(right, key=operator.itemgetter(1))
	left = sorted(left, key=operator.itemgetter(1))

	return np.array(left), np.array(right)

