import threading
import serial
import socket
from time import sleep
import numpy as np
from math import sqrt

class gps:
    def __init__(self):
        self.s = serial.Serial('/dev/tty_ublox', 4800, timeout=1)
        self.read_lock = threading.Lock()
        self.started = False
        self.lon = 0
        self.lat = 0
        self.fix = 0
        
    def start(self):
        if self.started:
            print('[!] GPS thread already started')
            return None
        self.started = True
        self.thread = threading.Thread(target=self.update, args=())
        self.thread.start()
        return self

    def update(self):
        while self.started:
            line = self.s.readline()  
            if line[1:6].decode() == "GNGSA":
                self.fix = int(line[9:10])
            
            if line[1:6].decode() == "GNGGA":
                line = line.decode()  
                if self.fix > 1:
                    with self.read_lock:
                        self.lat =  float(line[17:19]) + float(line[19:27])/60
                        self.lon =  float(line[31:33]) + float(line[33:41])/60
    def read(self):
        with self.read_lock:
            lat = self.lat
            lon = self.lon
            fix = self.fix
        return round(lat,6),round(lon,6),fix
    def stop(self):
        self.started = False
        self.thread.join()

    def __exit__(self, exec_type, exc_value, traceback):
        self.s.close()

        

def mapDist(lat, lon):
    earthRad = 6373.0
    zeroLat = 55.792978
    zeroLon = 12.509475

    zeroLat = np.deg2rad(zeroLat)
    zeroLon = np.deg2rad(zeroLon)

    lat = np.deg2rad(lat)
    lon = np.deg2rad(lon)

    y = lat-zeroLat
    x = (lon-zeroLon)*np.cos((lat+zeroLat)/2)

    distX = earthRad* x
    distY = earthRad* y

    return distX, distY 

def twoPtDist(lat1, lon1, lat2, lon2):
    earthRad = 6373.0

    lat1 = np.deg2rad(lat1)
    lon1 = np.deg2rad(lon1)
    lat2 = np.deg2rad(lat2)
    lon2 = np.deg2rad(lon2)

    y = lat2-lat1
    x = (lon2-lon1)*np.cos((lat1+lat2)/2)

    return earthRad*sqrt(x**2 + y**2)

def findPOI(poi,lat,lon):
    minDist = 1e9
    minIdx = None
    
    for i in range(len(poi)):
        dist = twoPtDist(lat, lon, poi[i,0], poi[i,1])
        if dist < minDist:
            minDist  = dist
            minIdx = i
    #print(minDist)
    if minDist < 0.005:
        return minIdx
    else: 
        return -1