import numpy as np
import cv2 as cv
import os



###########FUNCTIONS###########

def grabFrame(cap):
	ret, frame = cap.read()

	frame = cv.resize(frame,(width,height),interpolation = cv.INTER_AREA)

	return frame

def pointSelect(event, x, y, flags, param):
	global pt

	if event == cv.EVENT_LBUTTONDOWN:
		pt.append([x,y])

def autoExp(frame):
	meanSp = 125
	global exposure

	mean = np.mean(frame)

	if mean > meanSp+20:
		if exposure < 10 and exposure > 0:
			exposure -=1
		elif exposure > 0:
			exposure -= 2
		cmd = "v4l2-ctl -c exposure_absolute=" + str(exposure)
		os.system(cmd)
	elif mean < meanSp-20:
		if exposure < 10 and exposure > 0:
			exposure += 1
		elif exposure > 0:
			exposure += 2
		cmd = "v4l2-ctl -c exposure_absolute=" + str(exposure)
		os.system(cmd)

	return

def nothing(x):
	pass


###########INITIALIZATION############

os.system("v4l2-ctl -c focus_auto=0")
os.system("v4l2-ctl -c focus_absolute=0")
os.system("v4l2-ctl -c exposure_auto_priority=0")
os.system("v4l2-ctl -c exposure_auto=1")
os.system("v4l2-ctl -c exposure_absolute=20")

exposure = 200

cap = cv.VideoCapture(0)

cv.namedWindow('Frame',cv.WINDOW_NORMAL)
cv.resizeWindow('Frame',700,700)
cv.moveWindow('Frame',0,0)
cv.setMouseCallback('Frame',pointSelect)

height, width = 128,128
hCorr, wCorr = 428,210
warpMultiplier = 45
corrMat = np.eye(3)
corrMat[0,2] = corrMat[0,2] + wCorr
corrMat[1,2] = corrMat[1,2] + hCorr

ptTrue = np.array([[60.1,120.0],[0,120.0],[0,0],[60.1,0]])
pt = []

h = None

###########MAIN#############

while(True):

	frame = grabFrame(cap)
	frameWarped = frame.copy()

	autoExp(cv.cvtColor(frame,cv.COLOR_BGR2GRAY))
	

	for point in pt:
		cv.circle(frame,(point[0],point[1]),1,(0,0,255),-1)

	cv.line(frame,(64,0),(64,127),(0,0,255),1)
	cv.imshow('Frame',frame)

	if len(pt) == 4:
		cv.namedWindow('Homography',cv.WINDOW_NORMAL)
		cv.resizeWindow('Homography',400,400)
		cv.moveWindow('Homography',300,0)
		cv.createTrackbar('Height Correction','Homography',hCorr,1000,nothing)
		cv.createTrackbar('Width Correction','Homography',wCorr,1000,nothing)
		cv.createTrackbar('Multiplier','Homography',warpMultiplier,100,nothing)


		hCorr = cv.getTrackbarPos('Height Correction','Homography')
		wCorr = cv.getTrackbarPos('Width Correction','Homography')
		warpMultiplier = cv.getTrackbarPos('Multiplier','Homography')
		
		corrMat = np.eye(3)
		corrMat[0,2] = corrMat[0,2] + wCorr
		corrMat[1,2] = corrMat[1,2] + hCorr

		h,_ = cv.findHomography(np.array(pt),ptTrue)
		h = np.matmul(corrMat,h)

		frameWarped = cv.warpPerspective(frameWarped,h,(int(width*warpMultiplier/10),int(height*warpMultiplier/10)))
		frameWarped = cv.resize(frameWarped,(width,height))
		cv.line(frameWarped,(64,0),(64,127),(0,0,255),1)

		cv.imshow('Homography',frameWarped)
		

	key = cv.waitKey(1) & 0xFF

	if key == ord('r'):
		pt = []

	elif key == ord('q'): #13:  #13 = Enter pressed 
		break

if h is not None:
	print("Calculated Homography: ")
	print(h)
	#np.save('homography.npy',h)
	print("Corrections: ")
	print("Height: {0}\nWidth: {1}\nMultiplier: {2}".format(hCorr,wCorr,warpMultiplier/10))



##########RELEASE RESOURCES###########

cap.release()
cv.destroyAllWindows()