import numpy as np
import cv2 as cv
import neuralNetwork as nn
import cvLib
import time
import datetime
import sys
import socket
import os
import subprocess
import math
import gamepad
import queue
from autoExp import autoExp
from videocaptureasync import VideoCaptureAsync
from gps import gps, mapDist, findPOI


########GLOBAL FLAGS###########

debug = False
pause = False
logStart = None

odoLog = None
gpsLog = None
gpsGlobal = None
logCount = 0

ccw = True
eraseFlag = False
boxShown = False
printFlag = True
tiny = True
tidgemt = 0
tidcount = 0


for i in range(1,len(sys.argv)):
	if sys.argv[i] == '-debug':
		debug = True
	elif sys.argv[i] == '-cw':
		ccw = False
	elif sys.argv[i] == '-tiny':
		tiny = True
	elif sys.argv[i] == '-log':
		gpsGlobal = open(sys.argv[i+1], 'a')

###########FUNCTIONS###########

def grabFrame(cap):
	ret, frame = cap.read()

	frame = cv.resize(frame,(width,height),interpolation = cv.INTER_AREA)

	return frame

def getOdo(socket):
	
	idxx = -1
	idxy = -1
	idxth = -1

	cmd = 'eval $odox ; $odoy ; $odoth\r\n'
	socket.sendall(cmd.encode())

	while idxx < 0:

		data = socket.recv(1024)
		data = repr(data)

		idxx = data.find('odox')
		idxy = data.find('odoy')
		idxth = data.find('odoth')

	return float(data[idxx+7:idxx+15]), float(data[idxy+7:idxy+15]), float(data[idxth+7:idxth+15])

def mapCallback(event,x,y,flags,param):

	global eraseFlag, boxShown, poi, poiID, font

	if event == cv.EVENT_MOUSEMOVE:
		
		for i in range(len(poi)):
			distX, distY = mapDist(poi[i,0],poi[i,1])
			poiX = int((distX/mapWidth)*dtuMap.shape[1])
			poiY = int((abs(distY)/mapHeight)*628)

			if (poiX > x-5) and (poiX < x+5) and (poiY > y-5) and (poiY < y+5): 

				if poiID != i and not boxShown:
					if x <= 350:
						txt1 = 'Vel: {0} Edge: {1} Odo: {2}'.format(poi[i,2],poi[i,3],poi[i,4])
						txt2 = 'Stop: {0} Offset: {1} LookFwd: {2}'.format(poi[i,5],poi[i,6],poi[i,7])

						cv.rectangle(mapDrawn,(x,y-20),(x+150,y),(220,220,220),-1)
						cv.rectangle(mapDrawn,(x,y-20),(x+150,y),(0,0,0),1)
						cv.putText(mapDrawn,txt1,(x+3,y-12),font,0.25,(0,0,0),1,cv.LINE_AA)
						cv.putText(mapDrawn,txt2,(x+3,y-3),font,0.25,(0,0,0),1,cv.LINE_AA)
					else: 
						txt1 = "Vel: {0}".format(poi[i,2]) 
						txt2 = "Edge: {0}".format(poi[i,3]) 
						txt3 = "Odo: {0}".format(poi[i,4])
						txt4 = "Stop: {0}".format(poi[i,5])
						txt5 = "Offset: {0}".format(poi[i,6])
						txt6 = "LookFwd: {0}".format(poi[i,7])

						cv.rectangle(mapDrawn,(x,y-56),(x+58,y),(220,220,220),-1)
						cv.rectangle(mapDrawn,(x,y-56),(x+58,y),(0,0,0),1)

						cv.putText(mapDrawn,txt1,(x+3,y-49),font,0.25,(0,0,0),1,cv.LINE_AA)
						cv.putText(mapDrawn,txt2,(x+3,y-40),font,0.25,(0,0,0),1,cv.LINE_AA)
						cv.putText(mapDrawn,txt3,(x+3,y-31),font,0.25,(0,0,0),1,cv.LINE_AA)
						cv.putText(mapDrawn,txt4,(x+3,y-22),font,0.25,(0,0,0),1,cv.LINE_AA)
						cv.putText(mapDrawn,txt5,(x+3,y-13),font,0.25,(0,0,0),1,cv.LINE_AA)
						cv.putText(mapDrawn,txt6,(x+3,y-4),font,0.25,(0,0,0),1,cv.LINE_AA)

					boxShown = True

				poiID = i

				eraseFlag = True
				return
	
	boxShown = False
	eraseFlag = False	
	return

###########INITIALIZATION############

#Set camera params
print("Setting camera parameters...")
os.system("v4l2-ctl -c focus_auto=0")
os.system("v4l2-ctl -c focus_absolute=0")
os.system("v4l2-ctl -c exposure_auto_priority=0")
os.system("v4l2-ctl -c exposure_auto=1")
os.system("v4l2-ctl -c exposure_absolute=20")


print("Starting auto exposure thread...")
autoExp = autoExp()
autoExp.start()
print("Done.")

#Start MRC with no screen output
os.system("pkill mrc")
mrc = subprocess.Popen(["mrc", "-t1" ,"-v", "off"])
time.sleep(1)

#Connect to robot
print("Connecting to MRC, and resetting odometry...")
s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
server_address = ('localhost', 31001)
s.connect(server_address)
time.sleep(5)

#Setup the robot
cmd = 'set "xmlon" 1\r\n'
s.send(cmd.encode())
cmd = 'set "$odox" 0\r\n'
s.send(cmd.encode())
cmd = 'set "$odoy" 0\r\n'
s.send(cmd.encode())
cmd = 'set "$odoth" 0\r\n'
s.send(cmd.encode())
print("Done.")

#Load model
print("Loading model...") 
model = nn.loadModelMx('mx_model')
print("Done.")

if ccw:
	print("Reading points of interest from POI_CCW.csv... ")

	poi = np.genfromtxt('POI_CCW.csv', delimiter=',')
	poi = poi[1:,:-1]
else: 
	print("Reading points of interest from POI_CW.csv... ")

	poi = np.genfromtxt('POI_CW.csv', delimiter=',')
	poi = poi[1:,:-1]

print("Done.")

print("Starting GPS thread...")
#Start up GPS
os.system('fuser -k /dev/tty_ublox')
gps_reader = gps()
gps_reader.start() 
print("Done.")

#Load map and draw POI
print("Generating map...")

dtuMap = cv.imread("DTU_map.png",1)
mapWidth = 1.33
mapHeight = 1.65
distXOld = 0
distYOld = 0

font = cv.FONT_HERSHEY_SIMPLEX

cv.rectangle(dtuMap,(5,5),(130,45),(220,220,220),-1)
cv.rectangle(dtuMap,(5,5),(130,45),(0,0,0),2)
cv.putText(dtuMap,'X',(10,20), font, 0.5,(0,0,255),2,cv.LINE_AA)
cv.putText(dtuMap,': Manual Drive',(25,18), font, 0.4,(0,0,0),1,cv.LINE_AA)
cv.putText(dtuMap,'X',(10,40), font, 0.5,(0,255,0),2,cv.LINE_AA)
cv.putText(dtuMap,': Other',(25,38), font, 0.4,(0,0,0),1,cv.LINE_AA)

for i in range(len(poi)):
	distX, distY = mapDist(poi[i,0],poi[i,1])
	if poi[i,5] == 1:
		cv.putText(dtuMap,'X',(int((distX/mapWidth)*dtuMap.shape[1])-4,int((abs(distY)/mapHeight)*628)+5), font, 0.5,(0,0,255),2,cv.LINE_AA)
	else:
		cv.putText(dtuMap,'X',(int((distX/mapWidth)*dtuMap.shape[1])-4,int((abs(distY)/mapHeight)*628)+5), font, 0.5,(0,255,0),2,cv.LINE_AA)

mapDrawn = dtuMap.copy()
eraseFlag = False
poiID = -1
print("Done.")

print("Loading homography and parameters...")
#Load the homography, and parameters
h = np.load('homography.npy')
height, width = 128,128
warpMultiplier = 5
pixelsPerMeter = 20
lineAvg = 5
leftLine, rightLine = np.zeros((lineAvg,2)),np.zeros((lineAvg,2))
edgeToFollow = 1
edgeOffset = 0.8
lookFwd = 1.5	
velStart = 1
velTurbo = 1.5
camOffset = 1.08
lineCountL = 0
lineCountR = 0
interceptBottom = 63
pt = [63,63]
idxOld = -1
odothSaved = 0

startFlags = [velStart, edgeToFollow, -1, 0, edgeOffset, lookFwd, 0, 0, 0]
flags = startFlags
print("Done.")

print("Starting camera thread...")
if not debug:
	#Initialize the camera
	cap = VideoCaptureAsync(0)
	cap.start()
else:
	# If debug flag is set, use video file written here
	count = 0
	fourcc = cv.VideoWriter_fourcc(*'DIVX')
	cap = cv.VideoCapture('Dyrhaven.avi')
	maxCount = int(cap.get(cv.CAP_PROP_FRAME_COUNT))
	vidWrite = cv.VideoWriter('resultVid.avi',fourcc,3,(256,128))
print("Done.")

print("Starting gamepad thread...")
gamepad = gamepad.gamepad()
gamepad.start()
print("Done.")

print("Opening windows...")
#Initialize windows
cv.namedWindow('Frame',cv.WINDOW_NORMAL)
cv.moveWindow('Frame',0,0)
if tiny: 
	cv.resizeWindow('Frame',512,512)
else:
	cv.resizeWindow('Frame',375,375)

cv.namedWindow('Frame Warped',cv.WINDOW_NORMAL)
if tiny: 
	cv.moveWindow('Frame Warped',512,0)
	cv.resizeWindow('Frame Warped',512,512)
else:
	cv.moveWindow('Frame Warped',0,460)
	cv.resizeWindow('Frame Warped',375,375)


if gamepad.map:
	cv.namedWindow('Map',cv.WINDOW_NORMAL)
	cv.resizeWindow('Map',700,700)
	if tiny:
		cv.moveWindow('Map',235,0)
	else:
		cv.moveWindow('Map',470,0)
	cv.setMouseCallback('Map',mapCallback)
print("Done.")

###########MAIN LOOP#############
print("Starting main loop...")

while(not gamepad.exit):

	start = time.time()
	left, right = [],[]
	leftSlope, rightSlope = 0,0
	leftIntercept, rightIntercept = 0,0

	###Handle input from the gamepad###

	#Open/close the map window based on yellow button press
	if not gamepad.map:
		if cv.getWindowProperty('Map',0) == 0:
			cv.destroyWindow('Map')
	else:
		if cv.getWindowProperty('Map',0) == -1:
			cv.namedWindow('Map',cv.WINDOW_NORMAL)
			if tiny:
				cv.moveWindow('Map',162,0)
			else:
				cv.moveWindow('Map',470,0)
			cv.resizeWindow('Map',700,700)
			cv.setMouseCallback('Map',mapCallback)
			eraseFlag = False
	
	#Change color of path drawn on map based on manual drive or not
	if gamepad.color:
	  	mapColor = (0,0,255)
	else:	
		mapColor = (255,215,0)

	#Change speed based on green button press
	if gamepad.turbo:
		startFlags[0] = velTurbo
		startFlags[5] = 2.5
	else:
		startFlags[0] = velStart
		startFlags[5] = lookFwd
	#Bookkeeping for the map callback
	if (not eraseFlag) and gamepad.map:
		mapDrawn = dtuMap.copy()
	else:
		poiID = -1
		eraseFlag = True 

	###gamepad input handling ends###


	if debug:
		# set video frame
		cap.set(1,count)
		if not pause:
			if count+1 < maxCount:
				count += 1
				if count > 1:
					vis = np.concatenate((cv.addWeighted(frame,0.3,frameMasked,0.7,0),cv.addWeighted(frameWarped,0.3,frameMaskedWarped,0.7,0)),axis=1)
					vis = vis.reshape((128,256,3))
					vidWrite.write(vis)
					
			else: 
				gamepad.exit = True 
		else:
			# if paused, next/previous frame can be chosen with x/z and output images can be saved using s
			print("Frame no.: " + str(count))
			key = cv.waitKey(0)

			if key & 0xFF == ord('x'):
				count += 1
			elif key & 0xFF == ord('z'):
				if count > 0:
					count -= 1
			elif key & 0xFF == ord('s'):
				fname = 'Frame' + str(count) + ".png"
				testFrame = cv.addWeighted(testFrameWarped,0.3,testFrameMaskedWarped,0.7,0)
				
				cv.imwrite(fname,testFrame)

				fname = 'FrameRes' + str(count) + ".png"
				
				cv.imwrite(fname,cv.addWeighted(frameWarped,0.3,frameMaskedWarped,0.7,0))
			elif key & 0xFF == ord('p'):
				pause = False

	#Grab frame from camera and make copies for showing later on
	frame = grabFrame(cap)
	frameWarped = frame.copy()
	frameMasked = frame.copy()

	#Send the current frame to the exposure thread
	autoExp.frame = cv.cvtColor(frame,cv.COLOR_BGR2GRAY)
	
	
	#Generate segmentation of the path by passing the frame through the neural network
	mask = nn.forwardPassMx(cv.cvtColor(frame,cv.COLOR_BGR2RGB),model)
	mask = mask.astype('uint8')


	#Apply the mask to a copy of the original frame, and make a copy of the original frame to warp using the homography later
	frameMasked[mask == 255] = 255 
	frameMaskedWarped = frame.copy()

	#Warp the mask using the homography
	mask = cv.warpPerspective(mask,h,(int(width*warpMultiplier),int(height*warpMultiplier)))
	mask = cv.resize(mask,(width,height),interpolation = cv.INTER_AREA)

	# Find contours and remove contours with area < 100
	mask, cnt, _ = cv.findContours(mask, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE) 
	cnt = cvLib.removeSmallContours(cnt, 100)
	
	#Warp the image using the homography to achieve top-down view
	frameWarped = cv.warpPerspective(frameWarped,h,(int(width*warpMultiplier),int(height*warpMultiplier)))
	frameWarped = cv.resize(frameWarped,(width,height),interpolation = cv.INTER_AREA)
	testFrameWarped = frameWarped.copy()

	#warp the masked image as well
	frameMaskedWarped = cv.warpPerspective(frameMaskedWarped,h,(int(width*warpMultiplier),int(height*warpMultiplier)))
	frameMaskedWarped = cv.resize(frameMaskedWarped,(width,height),interpolation = cv.INTER_AREA)
	
	#apply the warped mask to the warped image
	frameMaskedWarped[mask == 255] = 255 
	testFrameMaskedWarped = frameMaskedWarped.copy()
	
	#If any contours remain after noise removal
	if len(cnt) > 0:

		#Generate approximate contour to reduce overall amount of data
		for i in range(len(cnt)):
			cnt[i] = cv.approxPolyDP(cnt[i], 2, True)
		
		#If exactly two contours remain in the image, try to merge them
		if len(cnt) == 2:
			cnt = cvLib.mergeContours(cnt)
		#If only 1 contour remains, generate a left and right edge of the path on the warped image
		if len(cnt) == 1:
			left, right = cvLib.splitLR(np.squeeze(cnt), mask, cv.cvtColor(frameWarped,cv.COLOR_BGR2GRAY))
			
	#If the right edge is more than 1 point long
	if len(right) > 1:

		#Remove points to close to other points
		right = cvLib.spacePoints(right)

		#If any points are still in the frame after spacing them, draw the result
		if len(right.shape) > 1:
			frameMaskedWarped = cvLib.drawEdge(frameMaskedWarped,right,(218,122,214),(128,0,128))

		#estimate line
		if len(right) > 1: 
			tempLine = np.polyfit(right[:,1],right[:,0]-int(flags[4]*pixelsPerMeter),1)

			if(rightLine[-1,1] == 0) or (abs(rightLine[1,0]-tempLine[0]) < 0.5) or lineCountR >= 5: 
				rightLine = np.append([tempLine],rightLine[0:-1],axis=0)
				lineCountR = 0
			else:
				rightLine = np.append([rightLine[0]],rightLine[0:-1],axis=0)
				lineCountR += 1

		#If the rolling average array is full, calculate the average slope and intercept of the lines in the array
		if rightLine[-1,1] != 0:
			rightSlope = np.sum(rightLine[:,0])/lineAvg
			rightIntercept = np.sum(rightLine[:,1])/lineAvg
								
	#If the left edge is more than 1 point long				
	if len(left) > 1:

		#Remove points to close to other points
		left = cvLib.spacePoints(left)

		#If any points are still in the frame after spacing them, draw the result
		if len(left.shape) > 1:
			frameMaskedWarped = cvLib.drawEdge(frameMaskedWarped,left,(255,215,0),(255,69,0))
		

		#estimate line
		if len(left) > 1:
			tempLine = np.polyfit(left[:,1],left[:,0]+int(flags[4]*pixelsPerMeter),1)

			if (leftLine[-1,1] == 0) or (abs(leftLine[1,0]-tempLine[0]) < 0.5 or lineCountL >= 5):
				leftLine = np.append([tempLine],leftLine[0:-1],axis=0)
				lineCountL = 0
			else:
				lineCountL +=1
				leftLine = np.append([leftLine[0]],leftLine[0:-1],axis=0)
		
		#If the rolling average array is full, calculate the average slope and intercept of the lines in the array
		if leftLine[-1,1] != 0:
			leftSlope = np.sum(leftLine[:,0])/lineAvg
			leftIntercept = np.sum(leftLine[:,1])/lineAvg

	
	#Read latitude and longitude of current position from GPS
	lat,lon,fix = gps_reader.read()	

	#if the GPS has a fix on the current position
	if fix > 1:

		#If a point of interest is close, return the idx of the POI in the list of POI's
		idx = findPOI(poi[:,0:2],lat,lon)

		#if no POI is close, set command flags to the standard values
		if idx  == -1:
			flags = startFlags
			printFlag = True
		#If a POI is close, set the command flags to the ones corresponding to the POI
		elif idx != idxOld:
			printFlag = True 	
			flags = poi[idx,2:]
			idxOld = idx

			#Print the command flags in the terminal
			if printFlag:
				print('Getting close to POI #{0}: {1}, setting flags to: {2}\n'.format(idx+1,poi[idx,0:2],flags))
				printFlag = False

		#Draw the current position on the map, and log the gps position
		if lat is not 0:
			distX, distY = mapDist(lat,lon)

			if (distX != distXOld) or (distY != distYOld):

				if distXOld != 0:

					if gpsGlobal is not None: 
						gpsGlobal.write('{0},\t{1},\t{2},\t{3} \n'.format(lat,lon,fix,int(gamepad.color)))

					cv.line(dtuMap,(int((distXOld/mapWidth)*dtuMap.shape[1]),int((abs(distYOld)/mapHeight)*628)),(int((distX/mapWidth)*dtuMap.shape[1]),int((abs(distY)/mapHeight)*628)),mapColor,4)
					

				distXOld = distX
				distYOld = distY 					
	
	#get the current odometry from the robot
	odox, odoy, odoth = getOdo(s)
	if flags[6]:
		odothSaved = odoth
		poi[idx,8] = 0

	#print('x: {0}\t y: {1}\t th: {2}\n'.format(odox,odoy,odoth))
	edgeToFollow = flags[1]

	#if right edge is selected, and it exists, find a point LookFwd distance away on the line		
	if edgeToFollow and (rightSlope != 0) and (rightIntercept != 0):
		pt = [(127-flags[5]*pixelsPerMeter)*rightSlope+rightIntercept, 127-flags[5]*pixelsPerMeter]
		interceptBottom = 127*rightSlope+rightIntercept
		
	#if the right edge does not exist, but the left does, use that instead
	elif (leftSlope != 0) and (leftIntercept != 0):
		pt = [(127-flags[5]*pixelsPerMeter)*leftSlope+leftIntercept, 127-flags[5]*pixelsPerMeter]
		interceptBottom = 127*leftSlope+leftIntercept
	#if left edge is selected, and it exists, find a point LookFwd distance away on the line
	if not edgeToFollow and (leftSlope != 0) and (leftIntercept != 0):
		pt = [(127-flags[5]*pixelsPerMeter)*leftSlope+leftIntercept, 127-flags[5]*pixelsPerMeter]
		interceptBottom = 127*leftSlope+leftIntercept
	#if the left edge does not exist, but the right does, use that instead
	elif (rightSlope != 0) and (rightIntercept != 0):
		pt = [(127-flags[5]*pixelsPerMeter)*rightSlope+rightIntercept, 127-flags[5]*pixelsPerMeter]
		interceptBottom = 127*rightSlope+rightIntercept

	#Draw lines on the image 
	cv.line(frameMaskedWarped, (int(rightIntercept),0), (int(127*rightSlope+rightIntercept),127), (128,0,128), 1, cv.LINE_AA)
	cv.line(frameMaskedWarped, (int(leftIntercept),0), (int(127*leftSlope+leftIntercept),127), (255,69,0), 1, cv.LINE_AA)

	# If lines are too noisy, set angle to go straight ahead
	if not flags[8] and ((len(left) > 1) and (len(right) > 1)) and((abs(leftLine[0,0]-rightLine[0,0]) > 1.5) or (interceptBottom > 127) or (interceptBottom < 0)):
		angle = 0
		cv.arrowedLine(frameMaskedWarped, (63,127), (63,127-int(flags[5]*pixelsPerMeter)), (0,69,255), 1, cv.LINE_AA, 0, 0.2)
	else:
		#calculate angle from the robots center to the point lookFwd distance away on the line chosen
		angle = np.rad2deg(math.atan2((127+(camOffset*pixelsPerMeter))-pt[1],63-pt[0]))-90
		cv.arrowedLine(frameMaskedWarped, (63,127), (int(pt[0]),int(pt[1])), (0,69,255), 1, cv.LINE_AA, 0, 0.2)

	th = np.rad2deg(odoth) - angle

	
	#If not in debug mode, tell the robot to drive
	if not debug:
		#flush the current commands on the robot 
		cmd = "flushcmds\r\n"
		s.send(cmd.encode()) 

		#calculate the velocity to send to the robot based on the turning angle
		vel = max(round(flags[0]*(1-abs(angle)/50),2),0.25)
		
		if abs(leftLine[0,0]-rightLine[0,0]) > 1.5:
			vel = 0.25
		#if the stop command flag is set, send the 'idle' command
		if flags[3]:
			cmd = "idle\r\n"
			s.send(cmd.encode())
		#if the robot should not drive using the odometry, send the standard command
		elif flags[2] == -1:
			cmd = "driveon " + str(odox) + " " + str(odoy) + " " + str(round(th,2)) + " @v" + str(vel) + ":($cmdtime>0.3)\r\n"
			s.send(cmd.encode())
		#drive the robot on the odometry
		else:
			print(np.rad2deg(odoth))
			print(np.rad2deg(odothSaved))
			print(flags[7])
			th = np.rad2deg(odoth)+ (flags[7]- np.rad2deg(odoth-odothSaved))
			print(th)
			cmd = "driveon " + str(odox) + " " + str(odoy) + " " + str(round(th,2)) + " @v1.0:($cmdtime>" + str(flags[2]) + ")\r\n"
			s.send(cmd.encode())

			time.sleep(flags[2])
			poi[idx,4] = -1	
	
	
	
	#Show the images to the user
	cv.imshow('Frame',cv.addWeighted(frame,0.3,frameMasked,0.7,0))
	cv.imshow('Frame Warped',cv.addWeighted(frameWarped,0.3,frameMaskedWarped,0.7,0))
	if gamepad.map:
		cv.imshow('Map',mapDrawn)

	#log data if needed
	if(gamepad.log):
		if logStart is None:
			logStart = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d_%H:%M:%S')

			print('Logging data in logs/log\ {0}'.format(logStart))

			os.system('mkdir logs/log\ {0}'.format(logStart))
			os.system('touch logs/log\ ' + logStart +'/odo.log')
			os.system('touch logs/log\ ' + logStart +'/gps.log')

			odoLog = open('logs/log ' + logStart + '/odo.log', 'w')
			odoLog.write('x, \ty, \t theta\n')

			gpsLog = open('logs/log ' + logStart + '/gps.log', 'w')
			gpsLog.write('latitude, \tlongitude, \t fix\n')

		odoLog.write('{0},\t {1},\t {2}\n'.format(odox,odoy,odoth))
		gpsLog.write('{0},\t {1},\t {2}\n'.format(lat,lon,fix))

		if logCount > 4:
			fname = 'logs/log ' + logStart + '/' + datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d_%H:%M:%S') + '.png'
			cv.imwrite(fname,frame)
			logCount = 0

		logCount += 1
	else: 
		if odoLog is not None:
			odoLog.close()
			odoLog = None
		if gpsLog is not None:
			gpsLog.close()
			gpsLog = None

		logStart = None
	
	#if 'q' is pressed quit, if 'p' is pressed, pause, if 's' is pressed, save the current position and an image to a temporary POI file
	key = cv.waitKey(1)

	if key & 0xFF == ord('q'): 
		break

	elif (key & 0xFF == ord('s')) or gamepad.saveImage:
		fname =  datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d-%H_%M_%S_%f') + '.png'
		string = str(lat) + ' , ' + str(lon) + ' , ' + fname + '\n'
		fname = 'lat-' + str(lat) + '-lon-' + str(lon) + '-' + fname
		cv.imwrite(fname,frame)
		print("Saved position and image.\n")

		gamepad.saveImage = False

	elif key & 0xFF == ord('p'):
		pause = True

	sys.stdout.write("Loop time: %.3f s \r" % (time.time()-start))
	sys.stdout.flush()

##########RELEASE RESOURCES###########
print("Releasing resources and stopping threads..")

cv.destroyAllWindows()
cv.imwrite('map_saved.png',dtuMap)

if debug:
	cap.release()
	vidWrite.release()
else:
	cap.stop()

s.send("flushcmds\r\n".encode()) 
s.send("idle\r\n".encode())
s.send("exit\r\n".encode())
s.close()
time.sleep(1)
mrc.kill()

if gamepad.log is not None:
	if odoLog is not None:
		odoLog.close()
	if gpsLog is not None: 
		gpsLog.close()

if gpsGlobal is not None:
	gpsGlobal.close()



gps_reader.stop()
autoExp.stop()
gamepad.stop()

print("Exiting.")

