#!/usr/bin/env python
# coding: utf-8

import keras

from keras.models import Model
from keras.layers import *
from keras import optimizers
from keras.callbacks import CSVLogger
from keras.preprocessing.image import ImageDataGenerator, image,img_to_array,array_to_img
import model.u_net as unet
import random
import time
from PIL import Image,ImageEnhance
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import os
import imageio
from keras.applications.vgg16 import preprocess_input
import cv2
import sys

#### GLOBAL VARIABLES ####
N_STEPS = 200000#len(os.listdir("/home/aut/s133910/train_img/img"))

val_steps = len(os.listdir("/home/aut/s133910/val_img/img"))

BATCH_SIZE = 6

im_size = 128
img_size = (im_size,im_size)

lr_init = 0.0002

csv_logger = CSVLogger('training.txt')

def pre_func(x):

    if  x.shape[0] == 3:
        x = array_to_img(x)
        x = ImageEnhance.Brightness(x).enhance(np.random.uniform(0.7, 1.3))
        x = img_to_array(x)
        x = preprocess_input(x)
    else:
        buf = np.random.uniform()
        if np.max(x) > 0:
            x = (x/np.max(x))*255
    return x 

def pre_func_val(x):

    if  x.shape[0] == 3:
        x = preprocess_input(x)
    return x 

data_gen_args = dict(preprocessing_function = pre_func,
                    rescale=1./255,
                    shear_range=0.3,
                    zoom_range=0.3,
                    horizontal_flip=True,
                    rotation_range = 7,
                    height_shift_range = 0.05,
                    width_shift_range = 0.05,
                    fill_mode='constant',
                    cval=0,
                    channel_shift_range = 10)


val_gen_args = dict(preprocessing_function = pre_func_val,
                    rescale=1./255)

train_datagen_img = ImageDataGenerator(**data_gen_args)
train_datagen_mask = ImageDataGenerator(**data_gen_args)


image_generator = train_datagen_img.flow_from_directory(
    '/home/aut/s133910/train_img/',
    target_size=img_size,
    batch_size=BATCH_SIZE,
    #color_mode = 'grayscale',
    seed = 1,
    shuffle = True,
    class_mode=None)

mask_generator = train_datagen_mask.flow_from_directory(
    '/home/aut/s133910/train_mask',
    target_size=img_size,
    batch_size=BATCH_SIZE,
    color_mode = 'grayscale',
    seed = 1,
    shuffle = True,
    class_mode=None)

val_datagen_img = ImageDataGenerator(**val_gen_args)
val_datagen_mask = ImageDataGenerator(**val_gen_args)

val_image_generator = val_datagen_img.flow_from_directory(
        '/home/aut/s133910/val_img/',
        target_size=img_size,
        batch_size=1,
        #color_mode = 'grayscale',
        seed = 1,
        shuffle = True,
        class_mode=None)

val_mask_generator = val_datagen_mask.flow_from_directory(
        '/home/aut/s133910/val_mask',
        target_size=img_size,
        batch_size=1,
        color_mode = 'grayscale',
        seed = 1,
        shuffle = True,
        class_mode=None)



train_generator = zip(image_generator, mask_generator)
val_generator = zip(val_image_generator, val_mask_generator)

model = unet.unet_base(input_shape=(3,im_size,im_size),learning_r = lr_init, batch_size = BATCH_SIZE)

callbacks = [csv_logger]

now = time.time()

model.fit_generator(
        train_generator,
        steps_per_epoch=N_STEPS/BATCH_SIZE,
        validation_data = val_generator,
        validation_steps = val_steps,
        callbacks = callbacks,
        epochs=36,verbose=1)

print("Elapsed time: " + str((time.time()-now)/60) + " m")
data_names, data_shapes = keras.models.save_mxnet_model(model= model, prefix='mx_model',epoch=0)
print(data_names)
print(data_shapes)
